import unittest

from . import system_information


class TestGenerateSystemInformation(unittest.TestCase):
    def test_generate_system_information(self):
        system_information.generate()


if __name__ == "__main__":
    unittest.main()



"""
This module implements Humbug's user consent mechanisms.
"""
import os
from typing import Callable, cast, Iterable, Optional, Sequence, Union

ConsentMechanism = Callable[[], bool]


class HumbugConsent:
    """
    HumbugConsent stores the client's consent settings.
    """

    BUGGER_OFF = "BUGGER_OFF"

    def __init__(self, *mechanisms: Union[bool, ConsentMechanism]) -> None:
        if not mechanisms:
            mechanisms = (False,)
        self._mechanisms = mechanisms
        self._bugger_off_mechanism = environment_variable_opt_out(self.BUGGER_OFF, yes)

    def check(self) -> bool:
        """
        Checks if all consent mechanisms signal the user's consent. If any of them signal False, returns
        False. Otherwise, returns True.
        """
        for mechanism in self._mechanisms:
            if mechanism is True:
                continue
            if mechanism is False:
                return False
            elif not cast(ConsentMechanism, mechanism)():
                return False
        # If the user has set BUGGER_OFF=yes then do not assume consent. Otherwise, at this point,
        # we can assume consent.
        return self._bugger_off_mechanism()


def environment_variable_opt_in(
    varname: str, opt_in_values: Sequence[str]
) -> ConsentMechanism:
    def mechanism() -> bool:
        if os.environ.get(varname) in opt_in_values:
            return True
        return False

    return mechanism


def environment_variable_opt_out(
    varname: str, opt_out_values: Sequence[str]
) -> ConsentMechanism:
    def mechanism() -> bool:
        if os.environ.get(varname) in opt_out_values:
            return False
        return True

    return mechanism


# yes and no are lists of values commonly used by environment variables to signify "yes" or "no"
yes = ["1", "t", "y", "T", "Y", "true", "yes", "True", "Yes", "TRUE", "YES"]
no = ["0", "f", "n", "F", "N", "false", "no", "False", "No", "FALSE", "NO"]


def prompt_user(
    prompt: str,
    accept_values: Iterable[str],
    reject_values: Iterable[str],
    retries: int = 0,
) -> ConsentMechanism:
    if accept_values is None:
        accept_values = yes
    if reject_values is None:
        reject_values = no

    def mechanism() -> bool:
        result: Optional[bool] = None
        attempts = 0
        while result is None:
            user_response = input(prompt)
            if user_response in accept_values:
                result = True
            elif user_response in reject_values:
                result = False
            else:
                if attempts >= retries:
                    result = False
                else:
                    attempts += 1
                    print("Invalid input: {}".format(user_response))
                    print(
                        "To accept, enter one of: {}".format(", ".join(accept_values))
                    )
                    print(
                        "To reject, enter one of: {}".format(", ".join(reject_values))
                    )

        return cast(bool, result)

    return mechanism



import os
import unittest
from unittest import mock

from . import consent


class TestHumbugConsent(unittest.TestCase):
    def setUp(self):
        os.environ["HUMBUG_TEST_OPT_IN_AFFIRMATIVE"] = "1"
        os.environ["HUMBUG_TEST_OPT_OUT_AFFIRMATIVE"] = "1"
        os.environ["HUMBUG_TEST_OPT_IN_NEGATIVE"] = "0"
        os.environ["HUMBUG_TEST_OPT_OUT_NEGATIVE"] = "0"
        self.ev_in_not_set = os.environ.get("HUMBUG_TEST_OPT_IN_NOT_SET")
        if self.ev_in_not_set is not None:
            del os.environ["HUMBUG_TEST_OPT_IN_NOT_SET"]
        self.ev_out_not_set = os.environ.get("HUMBUG_TEST_OPT_OUT_NOT_SET")
        if self.ev_out_not_set is not None:
            del os.environ["HUMBUG_TEST_OPT_OUT_NOT_SET"]

    def tearDown(self) -> None:
        del os.environ["HUMBUG_TEST_OPT_IN_AFFIRMATIVE"]
        del os.environ["HUMBUG_TEST_OPT_OUT_AFFIRMATIVE"]
        del os.environ["HUMBUG_TEST_OPT_IN_NEGATIVE"]
        del os.environ["HUMBUG_TEST_OPT_OUT_NEGATIVE"]
        if self.ev_in_not_set is not None:
            os.environ["HUMBUG_TEST_OPT_IN_NOT_SET"] = self.ev_in_not_set
        if self.ev_out_not_set is not None:
            os.environ["HUMBUG_TEST_OPT_OUT_NOT_SET"] = self.ev_out_not_set

    def test_base_consent_affirmative(self):
        consent_state = consent.HumbugConsent(True)
        self.assertTrue(consent_state.check())

    def test_base_consent_negative(self):
        consent_state = consent.HumbugConsent(False)
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_in_affirmative(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_in(
                "HUMBUG_TEST_OPT_IN_AFFIRMATIVE", ["true", "1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_envvar_opt_in_negative(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_in(
                "HUMBUG_TEST_OPT_IN_NEGATIVE", ["true", "1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_in_not_set(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_in(
                "HUMBUG_TEST_OPT_IN_NOT_SET", ["true", "1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_out_affirmative(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_out(
                "HUMBUG_TEST_OPT_OUT_AFFIRMATIVE", ["true", "1"]
            ),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_envvar_opt_out_negative(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_out(
                "HUMBUG_TEST_OPT_OUT_NEGATIVE", ["true", "1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_envvar_opt_out_not_set(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_out(
                "HUMBUG_TEST_OPT_OUT_NOT_SET", ["true", "1"]
            ),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_all_allow(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_in(
                "HUMBUG_TEST_OPT_IN_AFFIRMATIVE", ["1"]
            ),
            consent.environment_variable_opt_out("HUMBUG_TEST_OPT_OUT_NEGATIVE", ["1"]),
        )
        self.assertTrue(consent_state.check())

    def test_dynamic_consent_middle_reject_opt_in_var(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_in("HUMBUG_TEST_OPT_IN_NEGATIVE", ["1"]),
            consent.environment_variable_opt_out("HUMBUG_TEST_OPT_OUT_NEGATIVE", ["1"]),
        )
        self.assertFalse(consent_state.check())

    def test_dynamic_consent_middle_reject_opt_out_var(self):
        consent_state = consent.HumbugConsent(
            True,
            consent.environment_variable_opt_in(
                "HUMBUG_TEST_OPT_IN_AFFIRMATIVE", ["1"]
            ),
            consent.environment_variable_opt_out(
                "HUMBUG_TEST_OPT_OUT_AFFIRMATIVE", ["1"]
            ),
        )
        self.assertFalse(consent_state.check())

    @mock.patch.dict(os.environ, {consent.HumbugConsent.BUGGER_OFF: "true"})
    def test_bugger_off(self):
        consent_state = consent.HumbugConsent(True)
        self.assertFalse(consent_state.check())

    @mock.patch.dict(os.environ, {consent.HumbugConsent.BUGGER_OFF: "false"})
    def test_bugger_on(self):
        consent_state = consent.HumbugConsent(True)
        self.assertTrue(consent_state.check())

    # Testing input in this way was inspired by: https://stackoverflow.com/a/6272100/13659585
    @mock.patch.object(consent, "input")
    def test_prompt_user_accept(self, user_input):
        consent_checker = consent.HumbugConsent(
            consent.prompt_user("Accept? (yes/no)", ["yes"], ["no"])
        )
        user_input.return_value = "yes"
        self.assertTrue(consent_checker.check())
        self.assertEqual(user_input.call_count, 1)

    @mock.patch.object(consent, "input")
    def test_prompt_user_reject(self, user_input):
        consent_checker = consent.HumbugConsent(
            consent.prompt_user("Accept? (yes/no)", ["yes"], ["no"])
        )
        user_input.return_value = "no"
        self.assertFalse(consent_checker.check())
        self.assertEqual(user_input.call_count, 1)

    @mock.patch.object(consent, "print")
    @mock.patch.object(consent, "input")
    def test_prompt_user_invalid_responses(self, user_input, consent_print):
        retries = 2
        consent_checker = consent.HumbugConsent(
            consent.prompt_user("Accept? (yes/no)", ["yes"], ["no"], retries=retries)
        )
        user_input.return_value = "lol"
        self.assertFalse(consent_checker.check())
        self.assertEqual(user_input.call_count, retries + 1)
        self.assertEqual(consent_print.call_count, 3 * retries)


if __name__ == "__main__":
    unittest.main()



"""
This module implements all Humbug methods related to generating reports and publishing them to
Bugout knowledge bases.
"""
import atexit
import concurrent.futures
from dataclasses import dataclass, field
from enum import Enum
import os
import pkg_resources
import sys
import time
import traceback
from typing import List, Optional
import uuid

from bugout.app import Bugout

from .consent import HumbugConsent
from .system_information import (
    SystemInformation,
    generate as generate_system_information,
)


@dataclass
class Report:
    title: str
    content: str
    tags: List[str] = field(default_factory=list)


class Modes(Enum):
    DEFAULT = 0
    SYNCHRONOUS = 1


class Reporter:
    def __init__(
        self,
        name: str,
        consent: HumbugConsent,
        client_id: Optional[str] = None,
        session_id: Optional[str] = None,
        system_information: Optional[SystemInformation] = None,
        bugout_token: Optional[str] = None,
        bugout_journal_id: Optional[str] = None,
        timeout_seconds: int = 10,
        mode: Modes = Modes.DEFAULT,
    ):
        self.name = name
        self.consent = consent
        self.client_id = client_id
        if session_id is not None:
            self.session_id = session_id
        else:
            self.session_id = str(uuid.uuid4())
        if system_information is None:
            system_information = generate_system_information()
        self.system_information = system_information
        self.bugout = Bugout()
        self.bugout_token = bugout_token
        self.bugout_journal_id = bugout_journal_id
        self.timeout_seconds = timeout_seconds

        self.report_futures: List[concurrent.futures.Future] = []
        atexit.register(self.wait)

        self.executor: Optional[concurrent.futures.Executor] = None
        if mode == Modes.DEFAULT:
            self.executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=1, thread_name_prefix="humbug_reporter"
            )

        self.is_excepthook_set = False

    def wait(self) -> None:
        concurrent.futures.wait(
            self.report_futures, timeout=float(self.timeout_seconds)
        )
        if self.executor is not None:
            self.executor.shutdown()

    def system_tags(self) -> List[str]:
        tags = [
            "humbug",
            "source:{}".format(self.name),
            "os:{}".format(self.system_information.os),
            "arch:{}".format(self.system_information.machine),
            "python:{}".format(self.system_information.python_version_major),
            "python:{}.{}".format(
                self.system_information.python_version_major,
                self.system_information.python_version_minor,
            ),
            "python:{}".format(self.system_information.python_version),
            "session:{}".format(self.session_id),
        ]
        if self.client_id is not None:
            tags.append("client:{}".format(self.client_id))

        return tags

    def publish(self, report: Report, wait: bool = False) -> None:
        if not self.consent.check():
            return
        if self.bugout_token is None or self.bugout_journal_id is None:
            return

        try:
            report.tags = list(set(report.tags))
            if wait or self.executor is None:
                self.bugout.create_entry(
                    token=self.bugout_token,
                    journal_id=self.bugout_journal_id,
                    title=report.title,
                    content=report.content,
                    tags=report.tags,
                    timeout=self.timeout_seconds,
                )
            else:
                report_future = self.executor.submit(
                    self.bugout.create_entry,
                    token=self.bugout_token,
                    journal_id=self.bugout_journal_id,
                    title=report.title,
                    content=report.content,
                    tags=report.tags,
                    timeout=self.timeout_seconds,
                )
                self.report_futures.append(report_future)
        except:
            pass

    def custom_report(
        self,
        title: str,
        content: str,
        tags: Optional[List[str]] = None,
        publish: bool = True,
        wait: bool = False,
    ) -> Report:
        """
        Generates (and optionally publishes) a custom report in which the title, tags, and content
        are defined by the caller of this method.
        """
        if tags is None:
            tags = []
        report = Report(title=title, content=content, tags=tags)
        if publish:
            self.publish(report, wait=wait)
        return report

    def system_report(
        self, tags: Optional[List[str]] = None, publish: bool = True, wait: bool = False
    ) -> Report:
        title = "{}: System information".format(self.name)
        content = """### User timestamp
```
{user_time}
```

### OS
```
{os}
```

Release: `{os_release}`

### Processor
```
{machine}
```

### Python
```
{python_version}
```""".format(
            user_time=int(time.time()),
            os=self.system_information.os,
            os_release=self.system_information.os_release,
            machine=self.system_information.machine,
            python_version=self.system_information.python_version,
        )
        report = Report(title=title, content=content, tags=self.system_tags())
        if tags is not None:
            report.tags.extend(tags)
        report.tags.append("type:system")

        if publish:
            self.publish(report, wait=wait)

        return report

    def error_report(
        self,
        error: Exception,
        tags: Optional[List[str]] = None,
        publish: bool = True,
        wait: bool = False,
    ) -> Report:
        title = "{} - {}".format(self.name, type(error).__name__)
        error_content = """### User timestamp
```
{user_time}
```

### Exception summary
```
{error_summary}
```

### Traceback
```
{error_traceback}
```""".format(
            user_time=int(time.time()),
            error_summary=repr(error),
            error_traceback="".join(
                traceback.format_exception(
                    etype=type(error),
                    value=error,
                    tb=error.__traceback__,
                )
            ),
        )
        if tags is None:
            tags = []
        tags.append("type:error")
        tags.extend(self.system_tags())

        report = Report(title=title, content=error_content, tags=tags)

        if publish:
            self.publish(report, wait=wait)

        return report

    def env_report(
        self,
        title: Optional[str] = None,
        tags: Optional[List[str]] = None,
        publish: bool = True,
        wait: bool = False,
    ) -> Report:
        """
        Creates and optionally publishes a report containing the environment variables defined in
        the current process.
        """
        if title is None:
            title = "Environment variables"
        if tags is None:
            tags = []
        tags.append("type:env")

        env_vars = ["{}={}".format(key, value) for key, value in os.environ.items()]
        content = "```\n{}\n```".format("\n".join(env_vars))

        report = Report(title=title, content=content, tags=tags)
        if publish:
            self.publish(report, wait=wait)
        return report

    def packages_report(
        self,
        title: Optional[str] = None,
        tags: Optional[List[str]] = None,
        publish: bool = True,
        wait: bool = False,
    ) -> Report:
        """
        Creates and optionally publishes a report containing the packages (and versions of those
        packages) available in the current Python process.
        """
        if title is None:
            title = "Available packages"
        if tags is None:
            tags = []
        tags.append("type:dependencies")

        available_packages = [
            str(package_info) for package_info in pkg_resources.working_set
        ]
        content = "```\n{}\n```".format("\n".join(available_packages))
        report = Report(title, content, tags)
        if publish:
            self.publish(report, wait=wait)
        return report

    def compound_report(
        self,
        reports: List[Report],
        title: Optional[str] = None,
        tags: Optional[List[str]] = None,
        publish: bool = True,
        wait: bool = False,
    ) -> Report:
        if tags is None:
            tags = []
        for component in reports:
            tags.extend(component.tags)

        if title is None:
            title = "Composite report"

        content = "\n\n- - -\n\n".join(component.content for component in reports)

        report = Report(title=title, content=content, tags=tags)
        if publish:
            self.publish(report, wait=wait)
        return report

    def setup_excepthook(self, tags: Optional[List[str]] = None, publish: bool = True):
        """
        Adds error_report with python Exceptions.
        Only one excepthook will be added to stack, no matter how many
        times you call this method.

        Docs: https://docs.python.org/3/library/sys.html#sys.excepthook
        """
        if not self.is_excepthook_set:
            original_excepthook = sys.excepthook

            def _hook(exception_type, exception_instance, traceback):
                self.error_report(error=exception_instance, tags=tags, publish=publish)
                original_excepthook(exception_type, exception_instance, traceback)

            sys.excepthook = _hook

            self.is_excepthook_set = True

    def setup_notebook_excepthook(self, tags: Optional[List[str]] = None):
        """
        Excepthook for ipython, works with jupiter notebook.
        """
        ipython_shell = get_ipython()  # type: ignore
        old_showtraceback = ipython_shell.showtraceback

        def showtraceback(*args, **kwargs):
            _, exc_instance, _ = sys.exc_info()
            self.error_report(exc_instance, tags=tags, publish=True)
            old_showtraceback(*args, **kwargs)

        ipython_shell.showtraceback = showtraceback
        self.setup_excepthook(publish=True, tags=tags)



import unittest

from . import consent, report


class TestReporter(unittest.TestCase):
    def setUp(self):
        self.consent = consent.HumbugConsent(True)
        self.reporter = report.Reporter("TestReporter", self.consent)

    def test_system_report_successful(self):
        self.reporter.system_report(publish=False)

    def test_error_report_successful(self):
        error = Exception("This exception is for use in a Humbug Python test")
        self.reporter.error_report(error, publish=False)

    def test_custom_report_successful(self):
        title = "a"
        tags = ["b", "c", "d"]
        content = "e"
        report = self.reporter.custom_report(title, content, tags, publish=False)
        self.assertEqual(report.title, title)
        self.assertListEqual(report.tags, tags)
        self.assertEqual(report.content, content)

    def test_compound_report_successful(self):
        title = "a"
        tags = ["b", "c", "d"]
        content = "e"
        system_report = self.reporter.system_report(publish=False)
        custom_report = self.reporter.custom_report(title, content, tags, publish=False)
        title = "lol"
        compound_report = self.reporter.compound_report(
            [system_report, custom_report], title=title, tags=["d", "e"], publish=False
        )
        self.assertEqual(compound_report.title, title)
        self.assertSetEqual(
            set(compound_report.tags),
            set(self.reporter.system_tags() + tags + ["d", "e", "type:system"]),
        )

    def test_env_report_successful(self):
        tags = ["a", "b", "c"]
        env_report = self.reporter.env_report(tags=tags, publish=False)
        self.assertSetEqual(set(env_report.tags), set(tags + ["type:env"]))

    def test_packages_report_successful(self):
        tags = ["a", "b", "c"]
        pkg_report = self.reporter.packages_report(tags=tags, publish=False)
        self.assertSetEqual(set(pkg_report.tags), set(tags + ["type:dependencies"]))


if __name__ == "__main__":
    unittest.main()
